#include "raisim/World.hpp"
#include "raisim/object/Object.hpp"
#include "raisim/math.hpp"
#include "raisim/RaisimServer.hpp"
#include <math.h>

int main(int argc, char* argv[]) {
  auto binaryPath = raisim::Path::setFromArgv(argv[0]);

  raisim::World::setActivationKey("/home/weixuan/raisim/raisimExample/rsc/activation.raisim");
  raisim::World world("/home/weixuan/raisim/RaisimAerialInteraction/rsc/SimpleObject.xml");
  auto robot = world.addArticulatedSystem("/home/weixuan/raisim/RaisimAerialInteraction/rsc/simpleBiCopter.urdf","",{},raisim::COLLISION(63));
  auto robot_2 = world.addArticulatedSystem("/home/weixuan/raisim/RaisimAerialInteraction/rsc/simpleBiCopter2.urdf","",{},raisim::COLLISION(63));
  raisim::Vec<3> base_pos;
  base_pos[0] = -10;
  base_pos[1] = 0;
  base_pos[2] = 0.25;
  robot->setBasePos(base_pos);

  raisim::Vec<3> base_pos_2;
  base_pos_2[0] = 5;
  base_pos_2[1] = 0;
  base_pos_2[2] = 10;
  robot_2->setBasePos(base_pos_2);

  // raisim::World world;
  // auto robot = world.addArticulatedSystem("/home/weixuan/raisim/raisimSimpleAerialInteraction/rsc/laikago.urdf");

  // auto box = world.addBox(1, 1, 1, 1, "box2");
  // box->setPosition(0, 8,0.5);
  auto box = world.getObject("box2");
  Eigen::Vector3d lateral_force_box2;
  lateral_force_box2 << 20, 0 ,0;
  // auto ground = world.addGround(-2);
  world.setTimeStep(0.002);

  // auto box_id = box->getIndexInWorld();

  auto static_world_id = robot->getBodyIdx("static_world");
  auto base_link_id = robot->getBodyIdx("base_link");
  // auto wall_base_id = wall->getBodyIdx("base_link");
  // auto wall_virtual_link_id = wall->getBodyIdx("virtual_link");
  // auto plane_base_id = plane->getBodyIdx("base_link");
  // auto plane_virtual_link_id = plane->getBodyIdx("virtual_link");
  /// launch raisim server for visualization. Can be visualized using raisimUnity
  raisim::RaisimServer server(&world);
  server.launchServer();
  Eigen::Vector3d force, torque, input_torque, input_force;
  force << 1 ,1, 9.81;
  torque << 0.0,0.0, 0.0;

  double pitch_ref = 0;
  raisim::Vec<3> pos_ref;
  pos_ref[0] = 2.6;
  pos_ref[1] = 1.1; 
  pos_ref[2] = 1;

  double attitude_gain = 1;
  double angular_velocity_gain = 0.1;
  raisim::Vec<3> pos_p_gain, pos_d_gain;
  pos_p_gain[0] = 1;
  pos_p_gain[1] = 0;
  pos_p_gain[2] = 1;

  pos_d_gain[0] = 0.5;
  pos_d_gain[1] = 0.;
  pos_d_gain[2] = 0.1;



  double linear_velocity_x_gain = 1;
  double linear_velocity_z_gain = 1;

  // robot_ref_orientation << 1, 0, 0, 
  //                         0, 1, 0,
  //                         0, 0, 1;

  raisim::Mat<3,3> robot_orientation;
  // robot_orientation << 1, 0, 0,
  //                     0, 1, 0,
  //                     0, 0, 1;

  for (int i=0; i<100000; i++) {
    raisim::MSLEEP(2);
    raisim::Vec<3> box_pos, static_world_pos_B, static_world_pos_W, base_link_pos_B, base_link_pos_W, world_pos_B, world_pos_W, plane_pos_B, plane_pos_W;
    raisim::Vec<3> wall_base_pos_B, wall_base_pos_W, wall_virtual_link_pos_B, wall_virtual_link_pos_W;
    raisim::Vec<3> plane_base_pos_B, plane_base_pos_W, plane_virtual_link_pos_B, plane_virtual_link_pos_W;  
    raisim::Vec<3> robot_linear_velocity, robot_angular_velocity;
    // force_W << 0 ,0, 20;

    robot->getPosition(base_link_id, base_link_pos_B, base_link_pos_W);
    robot->getPosition(static_world_id, static_world_pos_B, static_world_pos_W);
    robot->getOrientation(base_link_id, robot_orientation);
    robot->getVelocity(base_link_id, robot_linear_velocity);
    robot->getAngularVelocity(base_link_id, robot_angular_velocity);
    double pitch = atan2(-robot_orientation[2], robot_orientation[0]);
    // std::cout << "robot_orientation[0]: " << robot_orientation[0] << std::endl;
    // std::cout << "robot_orientation[1]: " << robot_orientation[1] << std::endl;
    // std::cout << "robot_orientation[2]: " << robot_orientation[2] << std::endl;
    // std::cout << "robot_orientation[3]: " << robot_orientation[3] << std::endl;
    // std::cout << "robot_orientation[0,0]: " << robot_orientation[0] << std::endl;
    std::cout << "Pitch angle " << pitch << std::endl;

 

    // wall->getPosition(wall_base_id, wall_base_pos_B, wall_base_pos_W);
    // wall->getPosition(wall_virtual_link_id, wall_virtual_link_pos_B, wall_virtual_link_pos_W);

    // plane->getPosition(plane_base_id, plane_base_pos_B, plane_base_pos_W);
    // plane->getPosition(plane_virtual_link_id, plane_virtual_link_pos_B, plane_virtual_link_pos_W);

    // std::cout << " wall base_link_pos_B: " << wall_base_pos_W << std::endl;
    // std::cout << " wall wall_virtual_link_pos_W: " << wall_virtual_link_pos_W << std::endl;

    //  std::cout << "plane base_link_pos_B: " << plane_base_pos_W << std::endl;
    // std::cout << "plane wall_virtual_link_pos_W: " << plane_virtual_link_pos_W << std::endl;


    // std::cout << "base_link_pos_B: " << base_link_pos_B << std::endl;
    std::cout << "robot position: " << base_link_pos_W << std::endl;
    std::cout << "robot linear velocity: " << robot_linear_velocity << std::endl;
    // std::cout << "static_world_pos_W: " << static_world_pos_W << std::endl;
    std::cout << "robot orientation: " << robot_orientation << std::endl;
    // std::cout << "wall base link: " << world_pos_W << std::endl;
    // std::cout << "plane_pos_W: " << plane_pos_W << std::endl;

    //simulate disturbances, force disturbance to be simulated
    // robot->setExternalTorque(base_link_id, torque);
    box->setExternalForce(0, lateral_force_box2);

    // robot->setExternalForce(base_link_id, force);

    //simulate inputs
    // double torque_input = attitude_gain*(pitch_ref - pitch) + angular_velocity_gain*(0 - robot_angular_velocity[1]);
    double torque_input = attitude_gain*(pitch_ref - pitch);

    for (int i = 0; i < 3; ++i)
    {
      input_force[i] = pos_p_gain[i]*(pos_ref[i] - base_link_pos_W[i]) + pos_d_gain[i]*(0 - robot_linear_velocity[i]);
    }
    //gravity compensation
    input_force[2] += robot->getMass(base_link_id)*9.81;
    input_force[0] += 20;
    input_torque << 0, torque_input, 0;
    robot->setExternalTorque(base_link_id, input_torque);
    robot->setExternalForce(base_link_id, input_force);

    std::cout << "Input force: " << input_force << std::endl;

    // box->setExternalForce(box_id, force);
    // box->getPosition(box_id,box_pos);
    // std::cout << box_pos << std::endl;
    server.integrateWorldThreadSafe();
  }

  server.killServer();
  std::cout<<"end of simulation"<<std::endl;
}
